import re
import math
from decimal import Decimal
from datetime import date
from backports.cached_property import cached_property
from itertools import islice

from sqlalchemy import create_engine

from .logger import logger


class Controllers():

    FIELD_RE = re.compile('<([-a-z()_]+)$')
    PARAM_RE = re.compile(':([-a-z()_]+)$')

    def __init__(self, connection_string, max_rows, debug):
        self.connection_string = connection_string
        self.max_rows = max_rows
        self.debug = debug

    @cached_property
    def engine(self):
        return create_engine(self.connection_string, pool_size=20, max_overflow=0)

    def query_db_streaming(self, query_str, formatters):
        try:
            headers, formatters = self.parse_formatters(formatters)

            with self.engine.connect() as connection:
                logger.debug('executing %r', query_str)
                result = connection.execution_options(stream_results=True)\
                    .execute(query_str)
                yield headers
                yield from (
                    [f(row) for f in formatters]
                    for row in map(self.jsonable, map(dict, result))
                )
        except Exception:
            logger.exception('EXC')
            raise

    def query_db(self, query_str, num_rows, page_size, page):
        try:
            with self.engine.connect() as connection:
                count_query = "select count(1) from (%s) s" % query_str
                logger.debug('executing %r', count_query)
                count = connection.execute(count_query).fetchone()[0]
                logger.debug('count %r', count)

                num_rows = min(num_rows, self.max_rows, page_size)
                pages = math.ceil(count / num_rows)
                page = min(page, pages-1)
                page = max(page, 0)
                offset = page * page_size
                query = "select * from (%s) s limit %s offset %s" % (query_str, num_rows, offset)
                logger.debug('executing %r', query)
                result = connection.execute(query)
                rows = list(map(dict, islice(iter(result), 0, num_rows)))
                rows = [self.jsonable(row) for row in rows]
                logger.debug('rowcount %r', len(rows))
        except Exception as e:
            return {
                'success': False,
                'error': str(e)
            }
        return {
            'success': True,
            'total': count,
            'pages': pages,
            'page_size': num_rows,
            'page': page,
            'rows': rows,
        }

    def parse_formatters(self, formatters):
        _headers = []
        _formatters = []
        for h in formatters:
            field = self.FIELD_RE.findall(h)
            if len(field) == 1:
                field = field[0]
            else:
                field = None
            matches = self.PARAM_RE.findall(h)
            funcs = []
            while len(matches) > 0:
                mod = matches[0]
                h = h[:-(len(mod)+1)]
                funcs.append(self.formatter(mod))
                matches = self.PARAM_RE.findall(h)
            f = self.getter(field or h)
            for g in reversed(funcs):
                f = self.compose(f, g)
            k = self.wrapper(f)
            _formatters.append(k)
            _headers.append(h)
        return _headers, _formatters

    def wrapper(self, f):
        def _f(row):
            return f('', row)
        return _f

    def getter(self, h):
        hdr = h

        def _f(x, row):
            return row[hdr]
        return _f

    def compose(self, f, g):
        def _f(x, row):
            return g(f(x, row), row)
        return _f

    def formatter(self, mod):
        if mod == 'number':
            def _f(x, row):
                return str(x)
            return _f
        elif mod == 'yesno':
            def _f(x, row):
                return 'Yes' if x else 'No'  # TODO
            return _f
        elif mod == 'comma-separated':
            def _f(x, row):
                if x and isinstance(x, list):
                    return ','.join(x)
            return _f
        else:
            def _f(x, row):
                return str(x)
            return _f

    def jsonable(self, obj):
        if isinstance(obj, Decimal):
            return float(obj)
        if isinstance(obj, date):
            return obj.isoformat()
        if isinstance(obj, list):
            return [self.jsonable(x) for x in obj]
        if isinstance(obj, dict):
            return dict((k, self.jsonable(v)) for k, v in obj.items())
        return obj



import codecs
import os
import csv
import urllib
import tempfile
from io import StringIO

from flask import Blueprint, Response, request, send_file, abort
from flask_jsonpify import jsonpify

import xlsxwriter

from .controllers import Controllers
from .logger import logger, logging

MAX_ROWS = int(os.environ.get('APISQL__MAX_ROWS', 100))
CONNECTION_STRING = os.environ.get('APISQL__DATABASE_URL')


class APISQLBlueprint(Blueprint):

    def __init__(self, connection_string=CONNECTION_STRING, max_rows=MAX_ROWS, debug=False):
        super().__init__('apisql', 'apisql')
        self.controllers = Controllers(
            connection_string, max_rows, debug
        )
        if debug:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)
        self.max_rows = max_rows

        self.add_url_rule(
            '/query',
            'query',
            self.query,
            methods=['GET']
        )
        self.add_url_rule(
            '/download',
            'download',
            self.download,
            methods=['GET']
        )

    def query(self):
        results = dict(total=0, rows=[])
        if not self.detect_bot():
            try:
                num_rows = int(request.values.get('num_rows', self.max_rows))
                page_size = int(request.values.get('page_size', num_rows))
                page = int(request.values.get('page', 0))
            except Exception:
                abort(400)
            sql = request.values.get('query')
            try:
                sql = codecs.decode(sql.encode('ascii'), 'base64').decode('utf8')
            except Exception:
                pass
            results = self.controllers.query_db(sql, num_rows=num_rows, page_size=page_size, page=page)
        return jsonpify(results)

    def download(self):
        format = request.values.get('format', 'csv')

        file_name = request.values.get('filename')
        # Create a default value here in case this parameter is not provided
        if file_name is None:
            file_name = 'query-results'

        formatters = request.values.get('headers').split(';')

        if format not in ('csv', 'xlsx'):
            abort(400)

        if self.detect_bot():
            headers = {
                'Content-Type': 'text/csv',
                'Content-Disposition': 'attachment; filename=bot-detected.csv'
            }
            return Response('', content_type='text/csv', headers=headers)

        mime = {
            'csv': 'text/csv',
            'xlsx': 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
        }[format]

        sql = request.values.get('query')
        try:
            sql = codecs.decode(sql.encode('ascii'), 'base64').decode('utf8')
        except Exception:
            pass
        results = self.controllers.query_db_streaming(sql, formatters)

        if format == 'csv':
            def generate():
                buffer = StringIO()
                writer = csv.writer(buffer)
                for row in results:
                    writer.writerow(row)
                    pos = buffer.tell()
                    buffer.seek(0)
                    ret = buffer.read(pos)
                    buffer.seek(0)
                    yield ret

            # Encode the filename in utf-8 and url encoding
            file_name_utf8_encoded = file_name.encode('utf-8')
            file_name_url_encoded = urllib.parse.quote(file_name_utf8_encoded)

            headers = {
                'Content-Type': mime,
                'Content-Disposition': 'attachment; filename=' + file_name_url_encoded + '.csv'
            }
            return Response(generate(),
                            content_type='text/csv', headers=headers)
        if format == 'xlsx':
            with tempfile.NamedTemporaryFile(mode='w+b', suffix='.xslx') as out:
                try:
                    workbook = xlsxwriter.Workbook(out.name)
                    worksheet = workbook.add_worksheet()
                    for i, row in enumerate(results):
                        for j, v in enumerate(row):
                            if v is not None:
                                try:
                                    worksheet.write_number(i, j, float(v))
                                except ValueError:
                                    worksheet.write(i, j, str(v))
                finally:
                    workbook.close()
                return send_file(out.name, mimetype=mime, as_attachment=True, attachment_filename=file_name + '.xlsx')

    def detect_bot(self):
        if request.user_agent.browser in ('google', 'aol', 'baidu', 'bing', 'yahoo'):
            logger.info('Bot detected %s: %s', request.user_agent.string, request.user_agent.browser)
        elif any(x in request.user_agent.string.lower() for x in ('applebot', 'yandexbot', 'petalbot')):
            logger.info('Bot detected %s: %s', request.user_agent.string, request.user_agent.browser)
        else:
            return False
        return True
